@@ -17,9 +17,6 @@ import bisect # for binary search
 import stp.tools.gui as gui
 
 
-######################## Constructors ########################
-
-
 # The random number generator to be used
 # _RNG = np.random.default_rng(seed=0) # Seeded one for testing
 _RNG = np.random.default_rng()
@@ -28,9 +25,12 @@ _RNG = np.random.default_rng()
 MACHINE_EPSILON = np.finfo(float).eps
 
 
+######################## Constructors ########################
+
+
 def rand_p(n=2, zeros=0):
     """ Creates a random probability distribution, currently implemented only 
-        with the uniform sampling.
+        with uniform sampling.
     
         Kwargs:
             n (int): the dimension of the desired distribution
@@ -60,19 +60,24 @@ def rand_p(n=2, zeros=0):
     return p / np.sum(p)
 
 
-def rand_rate_matrix(n=2):
+def rand_rate_matrix(n=2, seed=None):
     """ Generates a random, time-independent, n x n rate matrix consisting of 
         probabilities per unit time. Column normalized.
     
         Kwargs:
             n (int): the number of states the matrix will correspond to. Relates to the dimensions of the matrix.
+
+            seed (None/int): the seed for sampling. Unseeded (uses module-level rng) if None.
     
         Returns:
             (np.ndarray): the rate matrix.
     
     """
+    # Use module-level rng if no seed is provided, else, seed one
+    rng = _RNG if seed is None else np.random.default_rng(seed=seed)
+
     # Generates a random n x n matrix
-    W = _RNG.random( (n, n) )
+    W = rng.random( (n, n) )
 
     np.fill_diagonal(W, 0)
     # Normalize the columns to sum to zero.
@@ -99,7 +104,7 @@ def rate_to_transition_matrix(W, time_step):
 
 
 def rand_transition_matrix(n=2, time_step=1.0):
-    """ Generates a random, time-independent, discrete-time, transition matrix
+    """ Generates a random, time-independent, discrete time, transition matrix
         by first generating a random rate matrix and then matrix exponentiating it to incorporate the time step as an additional parameter.
     
         Kwargs:
@@ -114,39 +119,32 @@ def rand_transition_matrix(n=2, time_step=1.0):
     return rate_to_transition_matrix(rand_rate_matrix(n), time_step)
 
 
-def self_assembly_rate_matrix(energy=1, c=1, M=1, T=1):
+def self_assembly_rate_matrix(alpha=1, c=1, M=1):
     """ Generates a self-assembly rate matrix for a 3-state system following:
         https://aip.scitation.org/doi/10.1063/1.3662140.
         
         Kwargs:
-            energy (float): the negative of the "optimally bound level of energy"
+            alpha (float/function): the energy/temperature coupled parameter as a float or a function which returns alpha as a function of time. In the latter case the returned rate matrix will be a function which provides a np.ndarray as a function of time. The energy is the negative of the "optimally bound level of energy".
 
-            c (float): "concentration-like variable"
+            c (float): "concentration-like variable".
 
-            M (int): the degeneracy of the misbound level
-
-            T (float/function): the temperature as a float or a function which returns the temperature as a function of time. In the latter case the returned rate matrix will be a function which provides a np.ndarray as a function of time.
-    
+            M (int): the degeneracy of the misbound level.
         Returns:
             (np.ndarray/function): the time-independent rate matrix as a numpy array in the case where the temperature is constant. Otherwise returns a function corresponding to the time-dependent rate matrix.
     
     """
-    if not isinstance(M, int):
-        print('The degeneracy of the misbound level, M, must be an integer.')
-        sys.exit()
+    if not callable(alpha) and alpha <= 0:
+        raise ValueError('Must provide a nonnegative alpha parameter.')
 
-    if not callable(T) and T <= 0:
-        print('Must provide a nonnegative temperature.')
-        sys.exit()
-
-    if callable(T):
+    if callable(alpha):
         # T is a time-dependent temperature
         # Provide the time-dependent rate matrix
-        return lambda t : self_assembly_rate_matrix(energy=energy, c=c, M=M, T=T(t))
+        return lambda t : self_assembly_rate_matrix(alpha=alpha(t), c=c, M=M)
+
+    # if not isinstance(M, int):
+    #     raise ValueError('The degeneracy of the misbound level, M, must be an integer.')
 
     ### Main body ###
-
-    alpha = np.exp(-energy / (2 * T))
     W = np.array([
         [   -c * (M + 1),   alpha,      alpha**2    ],
         [   c * M,          -alpha,     0           ],
@@ -156,35 +154,35 @@ def self_assembly_rate_matrix(energy=1, c=1, M=1, T=1):
     return W
 
 
-def self_assembly_transition_matrix(energy=1, c=1, M=1, T=1, time_step=1):
+def self_assembly_transition_matrix(alpha=1, c=1, M=1, time_step=1):
     """ Generates a self-assembly, discrete time, transition matrix for a 
         3-state system following: https://aip.scitation.org/doi/10.1063/1.3662140. Done assuming any external control parameter is fixed for the duration of the time_step. This matrix is step dependent, so conversions will be done to time to calculate the current temperature.
     
         Kwargs:
-            energy (float): the negative of the "optimally bound level of energy"
+            alpha (float/function): the energy/temperature coupled parameter as a float or a function which returns alpha as a function of time. In the latter case the returned transition matrix will be a function which provides a np.ndarray as a function of time. The energy is the negative of the "optimally bound level of energy". This external control parameter will be fixed for the duration of the time_step.
 
             c (float): "concentration-like variable"
 
             M (int): the degeneracy of the misbound level
 
-            T (float/function): the temperature as a float or a function which returns the temperature as a function of time. In the latter case the returned transition matrix will be a function which provides a np.ndarray as a function of time. This external control parameter will be fixed for the duration of the time_step.
-
             time_step (float): the time step
-    
-    
+        
         Returns:
             (np.ndarray/function): the transition matrix
     
     """
-    if callable(T):
+    if callable(alpha):
         # Convert n to time.
         # This also serves to fix the temperature between observations to change
             # the control parameter discretely.
-        old_T = T
-        T = lambda n : old_T(n * time_step)
+        R = lambda n : self_assembly_transition_matrix(
+            alpha=alpha(n * time_step), c=c, M=M,
+            time_step=time_step
+        )
+        return R
 
     return rate_to_transition_matrix(
-            self_assembly_rate_matrix(energy=energy, c=c, M=M, T=T),
+            self_assembly_rate_matrix(alpha=alpha, c=c, M=M),
             time_step
         )
 
@@ -258,13 +256,13 @@ def get_path_probability(R, p, path):
             (float): the probability of observing this path.
     
     """
-    # Check if R is written as a numpy matrix, if so, convert it to a function
+    # Check if R is time-independent, if so, convert it to a function
         # of the observation step that simply returns the constant matrix.
-    if isinstance(R, np.ndarray):
+    if not callable(R):
         matrix = R
         R = lambda n: matrix
 
-    # Comprehension to collect transition probabilities for vectored product
+    # Comprehension to collect transition probabilities for vectorized product
     total_jump_prob = np.array([
         R(i + 1)[path[i+1], path[i]]
         for i in range( len(path) - 1 )
@@ -311,35 +309,37 @@ def complete_path_space(n, path_length):
     return np.array(paths)
 
 
-def KMC(W, p, num_paths, final_time, time_step=1, discrete=True, degenerate_threshold=0.985):
+def KMC(W, p, num_paths, path_length, time_step=1, seed=None, _degenerate_threshold=0.985):
     """ A Rejection-free Kinetic Monte Carlo (KMC) algorithm for simulating 
-        the time evolution of a system, where some processes can occur with known rates W = W(t). From: https://en.wikipedia.org/wiki/Kinetic_Monte_Carlo.
+        the discrete time evolution of a system, where some processes can occur with known (continuous time) rates W = W(t). The discrete time dynamics will be computed using the continuous time rate matrix, W, by freezing the control parameter for moments of time, time_step. From: https://en.wikipedia.org/wiki/Kinetic_Monte_Carlo.
         
         Args:
-            W (np.ndarray/function): rate matrix for np.ndarray. If function is provided then W is the function of time that provides a rate matrix (W = W(t)) for each moment in time. Will be used to implement driven systems. If W is just a rate matrix then W is just the time-homogeneous rate matrix.
+            W (np.ndarray/function): rate matrix for np.ndarray. If function is provided then W is the function of time that provides a rate matrix (W = W(t)) for each moment in time. Will be used to implement driven systems. If W is just a rate matrix then W(t) is just the time-homogeneous rate matrix.
 
             p (np.ndarray): the initial marginal distribution.
 
             num_paths (int): the number of paths to sample.
 
-            final_time (float): the final time.
+            path_length (int): the length of each path.
     
         Kwargs:
             time_step (float): the interval between changing of rate matrix (changing of control parameter), and the delay between observations.
 
-            discrete (bool): True if the the simulation is for discrete time and hence the control parameter will be switched accordingly. If False the algorithm will be done for continuous time and will return the time series too.
+            seed (None/int): the seed for sampling. Unseeded (uses module-level rng) if None.
 
-            degenerate_threshold (float): the fraction of paths that we require to be unique. If not satisfied KMC will be run again with an increased threshold.
+            _degenerate_threshold (float): the fraction of paths that we require to be unique. If not satisfied KMC will be run again with an increased threshold.
     
         Returns:
             (np.ndarray): matrix of paths sampled via KMC.
     
     """
+
+
     #------------- Helper functions -------------#
 
 
     def _KMC_jump(cumulatives, state):
-        """ Helper function which takes cumulatives, a state, and generates 
+        """ Helper: takes cumulatives, a state, and generates 
             the next likely jump via the KMC algorithm and the time it takes to observe the jump. Useful for discretizing the KMC.
             
             Args:
@@ -354,131 +354,121 @@ def KMC(W, p, num_paths, final_time, time_step=1, discrete=True, degenerate_thre
         # https://en.wikipedia.org/wiki/Kinetic_Monte_Carlo.
         # The cumulative relevant to the current state
         relevant_cumulative = cumulatives[:, state]
-        u = _RNG.uniform(0, 1)
+        u = rng.uniform(0, 1)
         scaled_total_rate = u * cumulatives[-1, state]
         next_state = bisect.bisect_left(relevant_cumulative, scaled_total_rate)
-        u = _RNG.uniform(0, 1)
+        u = rng.uniform(0, 1)
         jump_time = -np.log(u) / relevant_cumulative[-1]
 
         return next_state, jump_time
 
 
-    def _KMC_discrete(W, p, num_paths, final_time, time_step=1):
+    #------------- Initialization -------------#
 
-        #------------- Initialization -------------#
+    # Use module-level rng if no seed is provided, else, seed one
+    rng = _RNG if seed is None else np.random.default_rng(seed=seed)
 
-        # Read off the size of the state space.
-        n = W(0).shape[0]
-        state_options = np.arange(n)
-        # Read off the path length for the full time series observations given 
-            # the final time and the delay between observations.
-        # For final_time, time_step, and path_length, the relationship is:
-        # time_step * (path_length - 1) = final_time, which implies that
-        path_length = int(final_time / time_step + 1)
-
-        # Initialize the matrix storing the paths
-        paths = np.full( (num_paths, path_length), -1, dtype='int64' )
-        # Set the first column of the matrix to be initial state weighted by p.
-        paths[:, 0] = _RNG.choice(state_options, p=p, size=num_paths)
-
-        # Fill each column
-        for k in range(1, path_length):
-            # This will find the current time of the control parameter to fix
-                # the rate matrix since this is discrete time.
-            current_time = k * time_step
-            # Fix the rate matrix
-            current_W = W(current_time)
-
-            #------------- Main algorithm -------------#
-
-            # We need to remove the normalization for the algorithm and just 
-                # look at the positive rates.
-            posW = current_W.copy()
-            np.fill_diagonal(posW, 0)
-            # Matrix of partial sums, last one being the total rate
-            # Each row corresponds to the same final index for the partial sum
-            # Each column corresponds to which rates were summed together
-            cumulatives = np.array( [ [ 
-                np.sum(posW[:i+1, x])
-                for i in range(n)
-            ]   for x in range(n)
-            ] ).T
-
-            # Run through the path matrix.
-            for path in paths:
-                current_state = path[k - 1]
-                # The state we are currently in is the last nonnegative 
-                    # element of our path list (since the path matrix is
-                    # initialized with -1)
-                # current_state = path[path >= 0][-1]
-                
-                total_jump_times = 0
-                # Prepare for the while loop
-                next_state = current_state
-                # Keep evolving the state until it fits what would be observed
-                    # in a discrete time process.
-                while total_jump_times < time_step:
-                    potential_jump, jump_time = _KMC_jump(cumulatives, next_state)
-                    
-                    # Note that if the next jump takes longer than our 
-                        # time_step we won't observe it. Since we will
-                        # make the measurement before this happens.
-                    if total_jump_times + jump_time >= time_step:
-                        break
-
-                    total_jump_times += jump_time
-                    next_state = potential_jump
-
-                path[k] = next_state
-
-        ### Degenerate paths check ###
-        # Check for degenerate paths and sample again if needed.
-        paths = np.unique(paths, axis=0)
-        if paths.shape[0] / num_paths < degenerate_threshold:
-            # Generate a rate matrix of perturbations
-            deltaW = rand_rate_matrix(n) / 100
-            Wprime = lambda t : W(t) + deltaW
-
-            # Run KMC again with an increased degeneracy threshold and 
-                # a perturbed rate matrix.
-            new_paths = KMC(Wprime, p, num_paths, final_time, time_step=time_step, discrete=discrete, degenerate_threshold=degenerate_threshold-10E-9)
-
-            # Combine the path data
-            paths = np.unique( np.vstack((paths, new_paths)), axis=0 )
-            
-        # Return only the requested number of paths (in case we have extras
-            # from previous degeneracies)
-        return paths[:num_paths]
-
-    #------------- Body -------------#
-
-    if isinstance(W, np.ndarray):
+    if not callable(W):
         # Hold the rate matrix in memory.
         constW = W
         # Let the time-dependent rate matrix just be the constant matrix.
         W = lambda t: constW
 
-    if discrete:
-        return _KMC_discrete(W, p, num_paths, final_time, time_step=time_step)
+    # Read off the size of the state space.
+    n = W(0).shape[0]
+    state_options = np.arange(n)
+    # Read off the final time for the full time series observations given 
+        # the path length and the delay between observations.
+    # For final_time, time_step, and path_length, the relationship is:
+    # time_step * (number of jumps) = final_time
+    final_time = time_step * (path_length - 1)
 
-    print('Continuous time KMC has not been implemented yet.')
-    print('Exiting...')
-    sys.exit()
+    # Initialize the matrix storing the paths
+    paths = np.full( (num_paths, path_length), -1, dtype='int64' )
+    # Set the first column of the matrix to be initial state weighted by p.
+    paths[:, 0] = rng.choice(state_options, p=p, size=num_paths)
+
+    # Fill each column
+    for k in range(1, path_length):
+        # This will find the current time of the control parameter to fix
+            # the rate matrix since this is discrete time.
+        current_time = k * time_step
+        # Fix the rate matrix
+        current_W = W(current_time)
+
+        #------------- Main algorithm -------------#
+
+        # We need to remove the normalization for the algorithm and just 
+            # look at the positive rates.
+        posW = current_W.copy()
+        np.fill_diagonal(posW, 0)
+        # Matrix of partial sums, last one being the total rate
+        # Each row corresponds to the same final index for the partial sum
+        # Each column corresponds to which rates were summed together
+        cumulatives = np.array( [ [ 
+            np.sum(posW[:i+1, x])
+            for i in range(n)
+        ]   for x in range(n)
+        ] ).T
+
+        # Run through the path matrix.
+        for path in paths:
+            current_state = path[k - 1]
+            # The state we are currently in is the last nonnegative 
+                # element of our path list (since the path matrix is
+                # initialized with -1)
+            # current_state = path[path >= 0][-1]
+            
+            total_jump_times = 0
+            # Prepare for the while loop
+            next_state = current_state
+            # Keep evolving the state until it fits what would be observed
+                # in a discrete time process.
+            while total_jump_times < time_step:
+                potential_jump, jump_time = _KMC_jump(cumulatives, next_state)
+                
+                # Note that if the next jump takes longer than our 
+                    # time_step we won't observe it. Since we will
+                    # make the measurement before this happens.
+                if total_jump_times + jump_time >= time_step:
+                    break
+
+                total_jump_times += jump_time
+                next_state = potential_jump
+
+            path[k] = next_state
+
+    ### Degenerate paths check ###
+    # Check for degenerate paths and sample again if needed.
+    paths = np.unique(paths, axis=0)
+    if paths.shape[0] / num_paths < _degenerate_threshold:
+        # Generate a rate matrix of perturbations
+        deltaW = rand_rate_matrix(n, seed=seed) / 100
+        Wprime = lambda t : W(t) + deltaW
+
+        # Run KMC again with an increased degeneracy threshold and 
+            # a perturbed rate matrix.
+        new_paths = KMC(Wprime, p, num_paths, path_length, time_step=time_step, seed=seed, _degenerate_threshold=_degenerate_threshold-10E-9)
+
+        # Combine the path data
+        paths = np.unique( np.vstack((paths, new_paths)), axis=0 )
+        
+    # Return only the requested number of paths (in case we have extra unique 
+        # ones from previous degeneracies)
+    return paths[:num_paths]
 
 
 def direct_sampling(R, p):
-    """ Samples the path space directly to reflect the dynamics given by the 
-        initial marginal and the transition matrix.
+    """ Not yet been implemented: Samples the path space directly to reflect 
+        the dynamics given by the initial marginal and the transition matrix.
         
         Args:
-            arg1 (arg1 type): arg1 description.
-    
-        Kwargs:
-            karg1 (arg1 type): arg1 description.
-    
-    
+            R (function/np.ndarray): the (possibly time-dependent) transition matrix.
+
+            p (np.ndarray): the marginal distribution.
+     
         Returns:
-            (None): none
+            (np.ndarray): the sampled portion of the path space.
     
     """
     print('Direct sampling has not been implemented yet.')
@@ -489,6 +479,18 @@ def direct_sampling(R, p):
 
 def main():
     print('stochastic.py')
+
+    W = self_assembly_transition_matrix(alpha=lambda t : np.exp(-65/2*(0.25*np.cos(0.01*t)+0.32+0.1*t)))
+    p = np.array([1 - 2 * MACHINE_EPSILON, MACHINE_EPSILON, MACHINE_EPSILON])
+
+    paths = KMC(W, p, 500, 20, seed=0)
+    paths2 = KMC(W, p, 500, 20, seed=0)
+    print( f'paths.shape: {paths.shape}' )
+    
+    print( f'paths: {paths}' )
+    print( f'paths - paths2: {paths - paths2}' )
+    
+    
     
     
 if __name__ == '__main__':
